@@ -1,5 +1,5 @@
 package System::Command;
-$System::Command::VERSION = '1.109';
+$System::Command::VERSION = '1.117';
 use warnings;
 use strict;
 use 5.006;
@@ -8,7 +8,7 @@ use Carp;
 use Cwd qw( cwd );
 use IO::Handle;
 use Symbol ();
-use Scalar::Util qw( blessed );
+use Scalar::Util qw( blessed reftype );
 use List::Util qw( reduce );
 use System::Command::Reaper;
 
@@ -44,19 +44,24 @@ sub import {
 }
 
 # a few simple accessors
-for my $attr (qw( pid stdin stdout stderr exit signal core options )) {
+{
     no strict 'refs';
-    *$attr = sub { return $_[0]{$attr} };
-}
-for my $attr (qw( cmdline )) {
-    no strict 'refs';
-    *$attr = sub { return @{ $_[0]{$attr} } };
+    for my $attr (qw( pid stdin stdout stderr options )) {
+        *$attr = sub { return $_[0]{$attr} };
+    }
+    for my $attr (qw( exit signal core )) {
+        no strict 'refs';
+        *$attr = sub { $_[0]->is_terminated(); return $_[0]{$attr} };
+    }
+    for my $attr (qw( cmdline )) {
+        *$attr = sub { return @{ $_[0]{$attr} } };
+    }
 }
 
 # REALLY PRIVATE FUNCTIONS
 # a sub-process spawning function
 my $_spawn = sub {
-    my (@cmd) = @_;
+    my ($o, @cmd) = @_;
     my $pid;
 
     # setup filehandles
@@ -95,7 +100,7 @@ my $_spawn = sub {
         pipe $err, $ERR or croak "errput pipe(): $!";
 
         # an extra pipe to communicate exec() failure
-        pipe my $stat_r, my $stat_w;
+        pipe my ( $stat_r, $stat_w );
 
         # create the child process
         $pid = fork;
@@ -125,7 +130,7 @@ my $_spawn = sub {
                 close $err;
 
                 # setup process group if possible
-                setpgrp 0, 0 if $Config{d_setpgrp};
+                setpgrp 0, 0 if $o->{setpgrp} && $Config{d_setpgrp};
 
                 # close $stat_w on exec
                 my $flags = fcntl( $stat_w, F_GETFD, 0 )
@@ -192,7 +197,7 @@ sub new {
     my ( $class, @cmd ) = @_;
 
     # split the args
-    my @o;
+    my @o = { setpgrp => 1 };
     @cmd = grep { !( ref eq 'HASH' ? push @o, $_ : 0 ) } @cmd;
 
     # merge the option hashes
@@ -222,13 +227,41 @@ sub new {
 
     # update the environment
     if ( exists $o->{env} ) {
+        croak "ENV variables cannot be empty strings on Win32"
+            if MSWin32 and grep { defined and !length } values %{ $o->{env} };
         @ENV{ keys %{ $o->{env} } } = values %{ $o->{env} };
         delete $ENV{$_}
             for grep { !defined $o->{env}{$_} } keys %{ $o->{env} };
     }
 
+    # interactive mode requested
+    if ( $o->{interactive} ) {
+        croak "Can't run command in interactive mode: not a terminal"
+          unless -t STDIN;
+
+        system { $cmd[0] } @cmd;
+
+        my $self = bless {
+            cmdline => [@cmd],
+            options => $o,
+            stdin   => IO::Handle->new,
+            stdout  => IO::Handle->new,
+            stderr  => IO::Handle->new,
+            exit    => $? >> 8,
+            signal  => $? & 127,
+            core    => $? & 128,
+        }, $class;
+
+        defined reftype( $o->{$_} )
+          and reftype( $o->{$_} ) eq 'SCALAR'
+          and ${ $o->{$_} } = $self->{$_}
+          for qw( exit signal core );
+
+        return $self;
+    }
+
     # start the command
-    my ( $pid, $in, $out, $err ) = eval { $_spawn->(@cmd); };
+    my ( $pid, $in, $out, $err ) = eval { $_spawn->( $o, @cmd ); };
 
     # FIXME - better check error conditions
     if ( !defined $pid ) {
@@ -266,7 +299,7 @@ sub new {
 
     # create the subprocess reaper and link the handles and command to it
     ${*$in} = ${*$out} = ${*$err} = $self->{reaper}    # typeglobs FTW
-        = System::Command::Reaper->new($self);
+      = System::Command::Reaper->new( $self, { trace => $trace, th => $th } );
 
     return $self;
 }
@@ -276,13 +309,73 @@ sub spawn {
     return @{ $class->new(@cmd) }{qw( pid stdin stdout stderr )};
 }
 
-# delegate those to the reaper
-sub is_terminated { $_[0]{reaper}->is_terminated() }
-sub close         { $_[0]{reaper}->close() }
+sub loop_on {
+    my $self = shift;
+
+    # handle options and defaults
+    my %args = (
+        stderr => sub { print STDERR shift },
+        @_
+    );
+    for my $which ( grep exists $args{$_}, qw( stdout stderr ) ) {
+        if ( $args{$which} ) {
+            croak "'$which' option must be a CODE reference"
+              if reftype $args{$which} ne 'CODE';
+        }
+        else {
+            delete $args{$which};
+        }
+    }
+
+    # create an object for the class method
+    if ( !ref $self ) {
+        die "'command' attribute required by loop_on when used as a class method"
+          if !exists $args{command};
+        $self = $self->new( @{ $args{command} } );
+    }
+
+    require IO::Select;
+    my $select = IO::Select->new( $self->stdout, $self->stderr );
+
+    local $/ = $args{input_record_separator}
+      if exists $args{input_record_separator};
+
+    # loop until end of streams
+    while ( my @ready = $select->can_read ) {
+        for my $fh (@ready) {
+            my $which = $fh == $self->stdout ? 'stdout' : 'stderr';
+            if ( defined( my $line = <$fh> ) ) {
+                my $ret = 1;
+                $ret = $args{$which}->($line)
+                  if exists $args{$which};
+                return if !$ret;
+            }
+            else {
+                $select->remove($fh);
+                $fh->close;
+            }
+        }
+    }
+
+    # success in the Unix sense
+    return $self->exit == 0;
+}
+
+# delegate those to the reaper (when there's one)
+sub is_terminated {
+    return $_[0]{options}{interactive}
+      ? 1
+      : $_[0]{reaper}->is_terminated();
+}
+
+sub close {
+    $_[0]{reaper}->close() unless $_[0]{options}{interactive};
+    return $_[0];
+}
 
 1;
 
-
+__END__
 
 =pod
 
@@ -290,10 +383,6 @@ sub close         { $_[0]{reaper}->close() }
 
 System::Command - Object for running system commands
 
-=head1 VERSION
-
-version 1.109
-
 =head1 SYNOPSIS
 
     use System::Command;
@@ -384,6 +473,30 @@ On some systems, some commands may close standard input on startup,
 which will cause a SIGPIPE when trying to write to it. This will raise
 an exception.
 
+=item C<interactive>
+
+If true, the command will actually be run using the L<perlfunc/system>
+builtin. If C<STDIN> is not a terminal, the constructor will die.
+
+Not reaper object will be created, and the C<stdin>, C<stdout> and
+C<stderr> filehandles will point to dummy closed handles. The C<exit>,
+C<signal> and C<core> attributes will be correctly set.
+
+(Added in version 1.114.)
+
+=item C<setpgrp>
+
+By default, the spawned process is made the leader of its own process
+group using C<setpgrp( 0, 0 )> (if possible). This enables sending a
+signal to the command and all its child processes at once:
+
+    # negative signal is sent to the process group
+    kill -SIGKILL, $cmd->pid;
+
+Setting the C<setpgrp> option to a false value disables this behaviour.
+
+(Added in version 1.110.)
+
 =item C<trace>
 
 The C<trace> option defines the trace settings for System::Command.
@@ -409,7 +522,7 @@ At trace level 2, the options values are shown:
     System::Command opt[12834]: fatal => {128 => 1,129 => 1}
     System::Command opt[12834]: git => "/usr/bin/git"
 
-Note: The C<fatal> and C<git> options in the example above is actually
+Note: The C<fatal> and C<git> options in the example above are actually
 used by L<Git::Repository> to determine the command to be run, and
 ignored by System::Command. References are dumped using L<Data::Dumper>.
 
@@ -421,13 +534,29 @@ At trace level 3, the content of the C<env> option is also listed:
 If the command cannot be spawned, the trace will show C<!> instead of
 the pid:
 
-    System::Command: ! - does-not-exist
+    System::Command cmd[!]: does-not-exist
+
+(Added in version 1.108.)
+
+=item exit
+
+=item core
+
+=item signal
+
+The above three options can be set to point to a reference to a scalar,
+which will be automatically updated when the command is terminated. See
+the L</Accessors> section for details about what the attributes of the
+same name mean.
+
+(Added in version 1.114.)
 
 =back
 
 The System::Command object returned by C<new()> has a number of
 attributes defined (see below).
 
+
 =head2 close
 
     $cmd->close;
@@ -435,6 +564,10 @@ attributes defined (see below).
 Close all pipes to the child process, collects exit status, etc.
 and defines a number of attributes (see below).
 
+Returns the invocant, so one can do things like:
+
+    my $exit = $cmd->close->exit;
+
 =head2 is_terminated
 
     if ( $cmd->is_terminated ) {...}
@@ -445,6 +578,7 @@ If the process was indeed terminated, collects exit status, etc.
 and defines the same attributes as C<close()>, but does B<not> close
 all pipes to the child process.
 
+
 =head2 spawn
 
     my ( $pid, $in, $out, $err ) = System::Command->spawn(@cmd);
@@ -453,6 +587,37 @@ This shortcut method calls C<new()> (and so accepts options in the same
 manner) and directly returns the C<pid>, C<stdin>, C<stdout> and C<stderr>
 attributes, in that order.
 
+(Added in version 1.01.)
+
+=head2 loop_on
+
+    $cmd->loop_on(
+        stdout => sub { ... },
+        stderr => sub { ... },
+    );
+
+This method calls the corresponding code references with each line
+produced on the standard output and errput of the command.
+
+If the C<stdout> or C<stderr> argument is not given, the default is to
+silently drop the data for C<stdout>, and to pass through (to STDERR)
+the data for C<stderr>. To prevent any processing, pass a false value
+to the parameter.
+
+For example, the following line will silently run the command to
+completion:
+
+    $cmd->loop_on( stderr => '' );
+
+The method blocks until the command is completed (or rather, until
+its output and errput handles have been closed), or until one of the
+callbacks returns a false value.
+
+The return value is true if the command exited with status 0, and false
+otherwise (i.e. the Unix traditional definition of success).
+
+(Added in version 1.117.)
+
 =head2 Accessors
 
 The attributes of a System::Command object are also accessible
@@ -462,27 +627,27 @@ The object returned by C<new()> will have the following attributes defined:
 
 =over 4
 
-=item cmdline()
+=item cmdline
 
 Return the command-line actually executed, as a list of strings.
 
-=item options()
+=item options
 
 The merged list of options used to run the command.
 
-=item pid()
+=item pid
 
 The PID of the underlying command.
 
-=item stdin()
+=item stdin
 
 A filehandle opened in write mode to the child process' standard input.
 
-=item stdout()
+=item stdout
 
 A filehandle opened in read mode to the child process' standard output.
 
-=item stderr()
+=item stderr
 
 A filehandle opened in read mode to the child process' standard error output.
 
@@ -498,24 +663,35 @@ C<$fh> is destroyed, the subprocess will be reaped, thus avoiding zombies.
 (L<System::Command::Reaper> undertakes this process.)
 
 After the call to C<close()> or after C<is_terminated()> returns true,
-the following attributes will be defined:
+the following attributes will be defined (note that the accessors
+always run C<is_terminated()>, to improve their chance of getting
+a value if the process just finished):
 
 =over 4
 
-=item exit()
+=item exit
 
 The exit status of the underlying command.
 
-=item core()
-
-A boolean value indicating if the command dumped core.
-
-=item signal()
+=item signal
 
 The signal, if any, that killed the command.
 
+=item core
+
+A boolean value indicating if the command dumped core.
+
 =back
 
+Even when not having a reference to the System::Command object any more,
+it's still possible to get the C<exit>, C<core> or C<signal> values,
+using the options of the same name:
+
+    my $fh = System::Command->new( @cmd, { exit => \my $exit } )->stdout;
+
+Once the command is terminated, the C<$exit> variable will contain the
+value that would have been returned by the C<exit()> method.
+
 =head1 CAVEAT EMPTOR
 
 Note that System::Command uses C<waitpid()> to catch the status
@@ -524,8 +700,8 @@ code (or any module you C<use>) does something like the following:
 
     local $SIG{CHLD} = 'IGNORE';    # reap child processes
 
-System::Command will not be able to capture the C<exit>, C<core>
-and C<signal> attributes. It will instead set all of them to the
+System::Command will not be able to capture the C<exit>, C<signal>
+and C<core> attributes. It will instead set all of them to the
 impossible value C<-1>, and display the warning
 C<Child process already reaped, check for a SIGCHLD handler>.
 
@@ -567,6 +743,12 @@ The L<System::Command::Reaper> class was added after the addition
 of Git::Repository::Command::Reaper in L<Git::Repository::Command> 1.11.
 It was later removed from L<System::Command> version 1.03, and brought
 back from the dead to deal with the zombie apocalypse in version 1.106.
+The idea of a reaper class comes from Vincent Pit.
+
+Thanks to Tim Bunce for using L<Git::Repository> and making many
+suggestions based on his use and needs. Most of them turned into
+improvement for System::Command instead, once we figured out that the
+more general feature idea really belonged there.
 
 =head1 BUGS
 
@@ -580,6 +762,7 @@ You can find documentation for this module with the perldoc command.
 
     perldoc System::Command
 
+
 You can also look for information at:
 
 =over 4
@@ -602,9 +785,10 @@ L<http://search.cpan.org/dist/System-Command/>
 
 =back
 
+
 =head1 COPYRIGHT
 
-Copyright 2010-2013 Philippe Bruhat (BooK).
+Copyright 2010-2015 Philippe Bruhat (BooK).
 
 =head1 LICENSE
 
@@ -616,9 +800,3 @@ See L<http://dev.perl.org/licenses/> for more information.
 
 =cut
 
-
-__END__
-
-# ABSTRACT: Object for running system commands
-
-
